"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _set = _interopRequireDefault(require("lodash/set"));

var _has = _interopRequireDefault(require("lodash/has"));

var _mongodb = require("mongodb");

var _escapeStringRegexp = _interopRequireDefault(require("escape-string-regexp"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _objectWithoutProperties(source, excluded) { if (source == null) return {}; var target = _objectWithoutPropertiesLoose(source, excluded); var key, i; if (Object.getOwnPropertySymbols) { var sourceSymbolKeys = Object.getOwnPropertySymbols(source); for (i = 0; i < sourceSymbolKeys.length; i++) { key = sourceSymbolKeys[i]; if (excluded.indexOf(key) >= 0) continue; if (!Object.prototype.propertyIsEnumerable.call(source, key)) continue; target[key] = source[key]; } } return target; }

function _objectWithoutPropertiesLoose(source, excluded) { if (source == null) return {}; var target = {}; var sourceKeys = Object.keys(source); var key, i; for (i = 0; i < sourceKeys.length; i++) { key = sourceKeys[i]; if (excluded.indexOf(key) >= 0) continue; target[key] = source[key]; } return target; }

class QueryBuilder {
  constructor({
    fieldResolvers = {},
    relationFields = []
  } = {}) {
    this.fieldResolvers = fieldResolvers;
    this.relationFields = relationFields;
    this.comparators = {};
    this.addComparator('not', (query, fieldName, value) => {
      (0, _set.default)(query, `${fieldName}.$ne`, value);
    });
    this.addComparator('notIn', (query, fieldName, value) => {
      (0, _set.default)(query, `${fieldName}.$nin`, value);
    });
    this.addComparator('in', (query, fieldName, value) => {
      (0, _set.default)(query, `${fieldName}.$in`, value);
    });
    this.addComparator('lt', (query, fieldName, value) => {
      (0, _set.default)(query, `${fieldName}.$lt`, value);
    });
    this.addComparator('lte', (query, fieldName, value) => {
      (0, _set.default)(query, `${fieldName}.$lte`, value);
    });
    this.addComparator('gt', (query, fieldName, value) => {
      (0, _set.default)(query, `${fieldName}.$gt`, value);
    });
    this.addComparator('gte', (query, fieldName, value) => {
      (0, _set.default)(query, `${fieldName}.$gte`, value);
    });
    this.addComparator('between', (query, fieldName, value) => {
      const [min, max] = value;
      (0, _set.default)(query, `${fieldName}.$gt`, min);
      (0, _set.default)(query, `${fieldName}.$lt`, max);
    });
    this.addComparator('betweene', (query, fieldName, value) => {
      const [min, max] = value;
      (0, _set.default)(query, `${fieldName}.$gte`, min);
      (0, _set.default)(query, `${fieldName}.$lte`, max);
    });
    this.addComparator('notContains', (query, fieldName, value) => {
      (0, _set.default)(query, `${fieldName}.$not.$regex`, `.*${(0, _escapeStringRegexp.default)(value)}.*`);
      (0, _set.default)(query, `${fieldName}.$not.$options`, 'i');
    });
    this.addComparator('contains', (query, fieldName, value) => {
      (0, _set.default)(query, `${fieldName}.$regex`, `.*${(0, _escapeStringRegexp.default)(value)}.*`);
      (0, _set.default)(query, `${fieldName}.$options`, 'i');
    });
    this.addComparator('regex', (query, fieldName, value) => {
      (0, _set.default)(query, `${fieldName}.$regex`, value);
      (0, _set.default)(query, `${fieldName}.$options`, 'i');
    });
    this.addComparator('startsWith', (query, fieldName, value) => {
      (0, _set.default)(query, `${fieldName}.$regex`, `^${(0, _escapeStringRegexp.default)(value)}.*`);
      (0, _set.default)(query, `${fieldName}.$options`, 'i');
    });
    this.addComparator('endsWith', (query, fieldName, value) => {
      (0, _set.default)(query, `${fieldName}.$regex`, `.*${(0, _escapeStringRegexp.default)(value)}$`);
      (0, _set.default)(query, `${fieldName}.$options`, 'i');
    });
  }

  addComparator(name, resolver) {
    this.comparators[name] = {
      matcher: new RegExp(`^(.*)_(${name})$`),
      resolver
    };
  }

  toField(field) {
    if (field === 'id') {
      return '_id';
    }

    return field;
  }

  castValue(value, field) {
    if (field === 'id' || field === '_id') {
      return Array.isArray(value) ? value.map(_mongodb.ObjectID) : (0, _mongodb.ObjectID)(value);
    }

    return value;
  }

  applyField(query, field, value) {
    const comparator = Object.keys(this.comparators).find(name => this.comparators[name].matcher.test(field));

    if (!comparator) {
      (0, _set.default)(query, this.toField(field), this.castValue(value, field));
    } else {
      const [, fieldName] = field.match(this.comparators[comparator].matcher);
      this.comparators[comparator].resolver(query, this.toField(fieldName), this.castValue(value, fieldName));
    }
  }

  toMongoQuery(where) {
    const query = {};
    Object.keys(where).forEach(field => {
      const value = where[field];
      const resolver = this.fieldResolvers[field];

      if (resolver) {
        resolver(query, value, where, this);
        return;
      }

      if (this.relationFields.includes(field)) {
        query[field] = this.toMongoQuery(value);
        return;
      }

      if (['and', 'or', 'nor'].includes(field.toLowerCase())) {
        (0, _set.default)(query, `$${field.toLowerCase()}`, value.map(subWhere => this.toMongoQuery(subWhere)));
        return;
      }

      this.applyField(query, field, value);
    });
    return query;
  }

  applyPagination(query, {
    skip,
    after,
    first,
    last,
    pivot,
    orderBy
  }) {
    if (pivot) {
      if (!orderBy) {
        (0, _set.default)(query, `query._id.${after ? '$gt' : '$lt'}`, pivot._id);
      } else {
        orderBy.forEach(({
          field,
          order
        }) => {
          const comparators = after ? {
            asc: '$gt',
            desc: '$lt'
          } : {
            asc: '$lt',
            desc: '$gt'
          };

          if ((0, _has.default)(pivot, field)) {
            (0, _set.default)(query, `query.${field}.${comparators[order.toLowerCase()]}`, pivot[field]);
          }
        });
      }
    }

    if (first) {
      (0, _set.default)(query, 'limit', parseInt(first, 10));
    }

    if (skip) {
      (0, _set.default)(query, 'skip', parseInt(skip, 10));
    }

    if (last) {
      throw new Error('Operator "last" is not supported!'); // set(query, 'skip', parseInt(skip, 10) + (total - last));
    }

    return query;
  }

  applySorting(query, {
    orderBy = []
  }) {
    if (!orderBy.length) {
      return query;
    }

    if (!query.orderBy) {
      (0, _set.default)(query, 'orderBy', {});
    }

    orderBy.forEach(({
      field,
      order
    }) => {
      Object.assign(query.orderBy, {
        [field]: order
      });
    });
    return query;
  }

  toQuery(_ref) {
    let {
      where = {}
    } = _ref,
        restArgs = _objectWithoutProperties(_ref, ["where"]);

    const query = {
      query: this.toMongoQuery(where)
    };
    this.applyPagination(query, restArgs);
    this.applySorting(query, restArgs);
    return query;
  }

}

var _default = QueryBuilder;
exports.default = _default;